package java.util.concurrent.atomic;

import java.util.Objects;
import java.lang.reflect.Modifier;
import java.security.PrivilegedActionException;
import sun.reflect.misc.ReflectUtil;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.lang.reflect.Field;
import sun.misc.Unsafe;
import java.util.function.BinaryOperator;
import java.util.function.UnaryOperator;
import sun.reflect.CallerSensitive;
import sun.reflect.Reflection;

public abstract class AtomicReferenceFieldUpdater<T, V>
{
    @CallerSensitive
    public static <U, W> AtomicReferenceFieldUpdater<U, W> newUpdater(Class<U> clazz, Class<W> clazz2, String s) {
        return new AtomicReferenceFieldUpdater.AtomicReferenceFieldUpdaterImpl(clazz, clazz2, s, Reflection.getCallerClass());
    }
    
    protected AtomicReferenceFieldUpdater() {
    }
    
    public abstract boolean compareAndSet(T p0, V p1, V p2);
    
    public abstract boolean weakCompareAndSet(T p0, V p1, V p2);
    
    public abstract void set(T p0, V p1);
    
    public abstract void lazySet(T p0, V p1);
    
    public abstract V get(T p0);
    
    public V getAndSet(T t, V v) {
        Object value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, v));
        return value;
    }
    
    public final V getAndUpdate(T t, UnaryOperator<V> unaryOperator) {
        Object value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, unaryOperator.apply(value)));
        return value;
    }
    
    public final V updateAndGet(T t, UnaryOperator<V> unaryOperator) {
        Object value;
        Object apply;
        do {
            value = this.get(t);
            apply = unaryOperator.apply(value);
        } while (!this.compareAndSet(t, value, apply));
        return apply;
    }
    
    public final V getAndAccumulate(T t, V v, BinaryOperator<V> binaryOperator) {
        Object value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, binaryOperator.apply(value, v)));
        return value;
    }
    
    public final V accumulateAndGet(T t, V v, BinaryOperator<V> binaryOperator) {
        Object value;
        Object apply;
        do {
            value = this.get(t);
            apply = binaryOperator.apply(value, v);
        } while (!this.compareAndSet(t, value, apply));
        return apply;
    }
    
    private static final class AtomicReferenceFieldUpdaterImpl<T, V> extends AtomicReferenceFieldUpdater<T, V>
    {
        private static final Unsafe U;
        private final long offset;
        private final Class<?> cclass;
        private final Class<T> tclass;
        private final Class<V> vclass;
        
        AtomicReferenceFieldUpdaterImpl(Class<T> tclass, Class<V> vclass, String s, Class<?> currentClass) {
            Field f;
            int modifiers;
            Class<Object> type;
            try {
                f = (Field)AccessController.doPrivileged(new PrivilegedExceptionAction<Field>() {
                    @Override
                    public Field run() throws NoSuchFieldException {
                        return tclass.getDeclaredField(s);
                    }
                });
                modifiers = f.getModifiers();
                ReflectUtil.ensureMemberAccess(currentClass, tclass, null, modifiers);
                ClassLoader classLoader = tclass.getClassLoader();
                ClassLoader classLoader2 = currentClass.getClassLoader();
                if (classLoader2 != null && classLoader2 != classLoader && (classLoader == null || !isAncestor(classLoader, classLoader2))) {
                    ReflectUtil.checkPackageAccess(tclass);
                }
                type = f.getType();
            }
            catch (PrivilegedActionException ex) {
                throw new RuntimeException(ex.getException());
            }
            catch (Exception cause) {
                throw new RuntimeException(cause);
            }
            if (vclass != type) {
                throw new ClassCastException();
            }
            if (vclass.isPrimitive()) {
                throw new IllegalArgumentException("Must be reference type");
            }
            if (!Modifier.isVolatile(modifiers)) {
                throw new IllegalArgumentException("Must be volatile type");
            }
            this.cclass = ((Modifier.isProtected(modifiers) && tclass.isAssignableFrom(currentClass) && !isSamePackage(tclass, currentClass)) ? currentClass : tclass);
            this.tclass = tclass;
            this.vclass = vclass;
            this.offset = AtomicReferenceFieldUpdater.AtomicReferenceFieldUpdaterImpl.U.objectFieldOffset(f);
            return;
        }
        
        private static boolean isAncestor(ClassLoader parent, ClassLoader classLoader) {
            do {
                parent = parent.getParent();
                if (classLoader == parent) {
                    return true;
                }
                continue;
            } while (parent != null);
            return false;
        }
        
        private static boolean isSamePackage(Class<?> clazz, Class<?> clazz2) {
            return clazz.getClassLoader() == clazz2.getClassLoader() && Objects.equals(getPackageName(clazz), getPackageName(clazz2));
        }
        
        private static String getPackageName(Class<?> clazz) {
            String name = clazz.getName();
            int lastIndex = name.lastIndexOf(46);
            return (lastIndex != -1) ? name.substring(0, lastIndex) : "";
        }
        
        private final void accessCheck(T t) {
            if (!this.cclass.isInstance(t)) {
                this.throwAccessCheckException(t);
            }
        }
        
        private final void throwAccessCheckException(T t) {
            if (this.cclass == this.tclass) {
                throw new ClassCastException();
            }
            throw new RuntimeException(new IllegalAccessException("Class " + this.cclass.getName() + " can not access a protected member of class " + this.tclass.getName() + " using an instance of " + t.getClass().getName()));
        }
        
        private final void valueCheck(V v) {
            if (v != null && !this.vclass.isInstance(v)) {
                throwCCE();
            }
        }
        
        static void throwCCE() {
            throw new ClassCastException();
        }
        
        @Override
        public final boolean compareAndSet(T o, V expected, V x) {
            this.accessCheck(o);
            this.valueCheck(x);
            return AtomicReferenceFieldUpdater.AtomicReferenceFieldUpdaterImpl.U.compareAndSwapObject(o, this.offset, expected, x);
        }
        
        @Override
        public final boolean weakCompareAndSet(T o, V expected, V x) {
            this.accessCheck(o);
            this.valueCheck(x);
            return AtomicReferenceFieldUpdater.AtomicReferenceFieldUpdaterImpl.U.compareAndSwapObject(o, this.offset, expected, x);
        }
        
        @Override
        public final void set(T o, V x) {
            this.accessCheck(o);
            this.valueCheck(x);
            AtomicReferenceFieldUpdater.AtomicReferenceFieldUpdaterImpl.U.putObjectVolatile(o, this.offset, x);
        }
        
        @Override
        public final void lazySet(T o, V x) {
            this.accessCheck(o);
            this.valueCheck(x);
            AtomicReferenceFieldUpdater.AtomicReferenceFieldUpdaterImpl.U.putOrderedObject(o, this.offset, x);
        }
        
        @Override
        public final V get(T o) {
            this.accessCheck(o);
            return AtomicReferenceFieldUpdater.AtomicReferenceFieldUpdaterImpl.U.getObjectVolatile(o, this.offset);
        }
        
        @Override
        public final V getAndSet(T o, V newValue) {
            this.accessCheck(o);
            this.valueCheck(newValue);
            return AtomicReferenceFieldUpdater.AtomicReferenceFieldUpdaterImpl.U.getAndSetObject(o, this.offset, newValue);
        }
        
        static {
            U = Unsafe.getUnsafe();
        }
    }
}
